// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"
	"os/signal"
	"path"
	"syscall"
	"time"

	"github.com/getsentry/sentry-go"
	gnatsd "github.com/nats-io/gnatsd/server"
	"github.com/slicebit/qb"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"xbus.io/go-xbus/v3/logsentry"
	"xbus.io/xbus/server"
	"xbus.io/xbus/version"
)

var serveCmd = &cobra.Command{
	Use:   "serve",
	Short: "Run Xbus servers",
	Long: `Run one or several of the Xbus servers.
The servers are:
- nats: A genuine nats server
- broker: Exchanges envelopes with actors
- control: Handle control API`,
	PersistentPreRun: preRunCheckCertCertDir,
	Run:              xbusServeRun,
}

func init() {
	RootCmd.AddCommand(serveCmd)
	flags := serveCmd.PersistentFlags()

	flags.String(
		"instance.name", "xbus",
		"Name of the xbus instance",
	)

	flags.Bool(
		"instance.production", false,
		"Indicate that the instance is in a production environment",
	)

	flags.String(
		"instance.logo", "img/logoXbus-white.png",
		"A logo url (can be a 'http(s)', 'file' or 'data' uri), or a webpath "+
			" relative to the http gateway /ui path",
	)

	flags.Bool(
		"debug-sql-queries", false,
		"Logs sql queries to debug. Very verbose",
	)

	flags.Bool(
		"debug-sql-bindings", false,
		"Logs sql bindings to debug. Very verbose",
	)

	flags.Bool(
		"nats", true,
		"Start a NATS server",
	)

	flags.Bool(
		"broker", true,
		"Start the broker service",
	)

	flags.Bool(
		"clientapi", true,
		"Start the clientapi service",
	)

	flags.Bool(
		"emitter", true,
		"Start the emitter service",
	)

	flags.Bool(
		"control", true,
		"Start the control service",
	)

	flags.Bool(
		"director", true,
		"Start the director service",
	)

	flags.Bool(
		"envelopestorage", true,
		"Start the envelopestorage service",
	)

	flags.Bool(
		"metrics", true,
		"Start the metrics service",
	)

	flags.Bool(
		"postmortem", true,
		"Start the postmortem service",
	)

	flags.Bool(
		"processstate", true,
		"Start the processstate service",
	)

	flags.Bool(
		"registration", true,
		"Start the registration service",
	)

	flags.Bool(
		"sentinel", true,
		"Start the sentinel service",
	)

	flags.String(
		"gnatsd.http_host", "",
		"User host for http monitoring",
	)
	flags.Int(
		"gnatsd.http_port", 0,
		"User port for http monitoring",
	)
	flags.Int(
		"gnatsd.https_port", 0,
		"User port for https monitoring",
	)

	flags.String(
		"sentry.dsn", "",
		"DSN of a sentry instance",
	)

	flags.String(
		"sentry.environment", "",
		"DSN of a sentry instance",
	)

	flags.String(
		"metrics-http-host", "",
		"User host for http metrics (prometheus)",
	)

	flags.Int(
		"metrics-http-port", 0,
		"User port for http metrics (prometheus)",
	)

	flags.Bool(
		"metrics-merge-global", true,
		"Merge the bus-wide metrics (/metrics/global) into the xbusd instance metrics (/metrics)",
	)

	flags.Duration(
		"metrics-delay-hint", 15*time.Second,
		"Optimize metrics gathering for the given delay between collections",
	)

	viper.SetDefault("instance.name", "xbus")
	must(viper.BindPFlag("instance.name", flags.Lookup("instance.name")))

	viper.SetDefault("instance.production", false)
	must(viper.BindPFlag("instance.production", flags.Lookup("instance.production")))

	viper.SetDefault("instance.logo", "img/logoXbus-white.png")
	must(viper.BindPFlag("instance.logo", flags.Lookup("instance.logo")))

	viper.SetDefault("debug-sql-queries", false)
	must(viper.BindPFlag("debug-sql-queries", flags.Lookup("debug-sql-queries")))

	viper.SetDefault("debug-sql-bindings", false)
	must(viper.BindPFlag("debug-sql-bindings", flags.Lookup("debug-sql-bindings")))

	viper.SetDefault("serve.nats", true)
	must(viper.BindPFlag("serve.nats", flags.Lookup("nats")))

	viper.SetDefault("serve.broker", true)
	must(viper.BindPFlag("serve.broker", flags.Lookup("broker")))

	viper.SetDefault("serve.clientapi", true)
	must(viper.BindPFlag("serve.clientapi", flags.Lookup("clientapi")))

	viper.SetDefault("serve.control", true)
	must(viper.BindPFlag("serve.control", flags.Lookup("control")))

	viper.SetDefault("serve.director", true)
	must(viper.BindPFlag("serve.director", flags.Lookup("director")))

	viper.SetDefault("serve.envelopestorage", true)
	must(viper.BindPFlag("serve.envelopestorage", flags.Lookup("envelopestorage")))

	viper.SetDefault("serve.metrics", true)
	must(viper.BindPFlag("serve.metrics", flags.Lookup("metrics")))

	viper.SetDefault("serve.postmortem", true)
	must(viper.BindPFlag("serve.postmortem", flags.Lookup("postmortem")))

	viper.SetDefault("serve.processstate", true)
	must(viper.BindPFlag("serve.processstate", flags.Lookup("processstate")))

	viper.SetDefault("serve.registration", true)
	must(viper.BindPFlag("serve.registration", flags.Lookup("registration")))

	viper.SetDefault("serve.sentinel", true)
	must(viper.BindPFlag("serve.sentinel", flags.Lookup("sentinel")))

	must(viper.BindPFlag("gnatsd.http_host", flags.Lookup("gnatsd.http_host")))
	must(viper.BindPFlag("gnatsd.http_port", flags.Lookup("gnatsd.http_port")))
	must(viper.BindPFlag("gnatsd.https_port", flags.Lookup("gnatsd.https_port")))

	must(viper.BindPFlag("sentry.dsn", flags.Lookup("sentry.dsn")))
	must(viper.BindPFlag("sentry.environment", flags.Lookup("sentry.environment")))
	must(viper.BindPFlag("metrics.http-host", flags.Lookup("metrics-http-host")))
	must(viper.BindPFlag("metrics.http-port", flags.Lookup("metrics-http-port")))
	viper.SetDefault("metrics.merge-global", true)
	must(viper.BindPFlag("metrics.merge-global", flags.Lookup("metrics-merge-global")))
	viper.SetDefault("metrics.delay-hint", time.Second*15)
	must(viper.BindPFlag("metrics.delay-hint", flags.Lookup("metrics-delay-hint")))
}

func xbusServeRun(cmd *cobra.Command, args []string) {
	certDir := viper.GetString("cert-dir")

	logFlags := qb.LDefault
	if viper.GetBool("debug-sql-queries") {
		logFlags |= qb.LQuery
	}
	if viper.GetBool("debug-sql-bindings") {
		logFlags |= qb.LBindings
	}

	instance := server.InstanceProperties{
		Name:       viper.GetString("instance.name"),
		Production: viper.GetBool("instance.production"),
		Logo:       viper.GetString("instance.logo"),
	}

	var sentryHub *sentry.Hub
	if dsn := viper.GetString("sentry.dsn"); dsn != "" {
		sentryClient, err := sentry.NewClient(sentry.ClientOptions{
			Dsn:         viper.GetString("sentry.dsn"),
			Release:     version.Version,
			Environment: viper.GetString("sentry.environment"),
		})
		if err != nil {
			Logger.Zero().Err(err).Msg("Could not setup sentry")
		}
		scope := sentry.NewScope()
		scope.SetTags(map[string]string{
			"xbus.instance":   instance.Name,
			"xbus.production": fmt.Sprintf("%t", instance.Production),
		})
		sentryHub = sentry.NewHub(sentryClient, scope)
		// TODO initLogger w. sentry middleware
		initLogger(true, logsentry.GetSentryOutputMiddleware(sentryHub))
	}

	opts := server.Options{
		Instance:                       instance,
		DatabaseDSN:                    viper.GetString("database.dsn"),
		DatabaseMaxConnections:         viper.GetInt("database.max-connections"),
		ProcessStorageDBMaxConnections: viper.GetInt("storage.process.db.max-connections"),
		DatabaseLogFlags:               logFlags,
		RootCA:                         path.Join(certDir, "ServerCA.crt"),
		ServerCert:                     path.Join(certDir, "ServerCA.crt"),
		ServerKey:                      path.Join(certDir, "ServerCA.key"),
		ClientCA:                       path.Join(certDir, "ClientCA.crt"),
		ClientCAKey:                    path.Join(certDir, "ClientCA.key"),
		TLSTimeout:                     viper.GetFloat64("tls.timeout"),
		MetricsHTTPHost:                viper.GetString("metrics.http-host"),
		MetricsHTTPPort:                viper.GetInt("metrics.http-port"),
		MetricsMergeGlobal:             viper.GetBool("metrics.merge-global"),
		MetricsDelayHint:               viper.GetDuration("metrics.delay-hint"),
		NATSHost:                       viper.GetString("nats.host"),
		NATSPort:                       viper.GetInt("nats.port"),
		NATSDebug:                      viper.GetBool("debug"),
		NATSTrace:                      viper.GetBool("trace"),
		Sentry:                         sentryHub,
		Logger:                         Logger,
		ServeNATS:                      viper.GetBool("serve.nats"),
		ServeBroker:                    viper.GetBool("serve.broker"),
		ServeClientAPI:                 viper.GetBool("serve.clientapi"),
		ServeControl:                   viper.GetBool("serve.control"),
		ServeDirector:                  viper.GetBool("serve.director"),
		ServeEnvelopeStorage:           viper.GetBool("serve.envelopestorage"),
		ServeMetrics:                   viper.GetBool("serve.metrics"),
		ServePostmortem:                viper.GetBool("serve.postmortem"),
		ServeProcessState:              viper.GetBool("serve.processstate"),
		ServeRegistration:              viper.GetBool("serve.registration"),
		ServeSentinel:                  viper.GetBool("serve.sentinel"),
		GnatsdOptions: gnatsd.Options{
			HTTPHost:  viper.GetString("gnatsd.http_host"),
			HTTPPort:  viper.GetInt("gnatsd.http_port"),
			HTTPSPort: viper.GetInt("gnatsd.https_port"),
		},
	}

	s, err := server.NewServer(opts)
	if err != nil {
		Logger.Errorf("Could not setup the server: %s", err)
		os.Exit(1)
	}

	if err := s.Startup(); err != nil {
		Logger.Errorf("Could not start the server: %s", err)
		os.Exit(1)
	}

	Logger.Noticef("Server started, now waiting for a signal")

	// TODO do something a little more elaborate so that interrupt during startup
	// is gracefully handlded
	signals := make(chan os.Signal, 1)
	signal.Notify(signals, os.Interrupt, syscall.SIGTERM)

	sig := <-signals

	Logger.Noticef("Signal '%s' caught, shutting down", sig)

	s.Shutdown()
}
