package metrics

import (
	"context"
	"sync"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
)

func NewSlowCollector(
	ctx context.Context, name string, c prometheus.Collector,
	delayHint time.Duration, log zerolog.Logger,
) *SlowCollector {
	sc := SlowCollector{
		Collector: c,
		metricsC:  make(chan []prometheus.Metric, 1),
		delayHint: delayHint,
		log:       log.With().Str("slow-collector", name).Logger(),
	}
	sc.ctx, sc.stop = context.WithCancel(ctx)
	sc.wg.Add(1)
	go sc.run()
	return &sc
}

type SlowCollector struct {
	ctx       context.Context
	Collector prometheus.Collector

	stop      func()
	wg        sync.WaitGroup
	metricsC  chan []prometheus.Metric
	delayHint time.Duration
	log       zerolog.Logger
}

// Describe implements the prometheus.Collector interface.
func (c *SlowCollector) Describe(ch chan<- *prometheus.Desc) {
	c.Collector.Describe(ch)
}

// Collect implements the prometheus.Collector interface.
func (c *SlowCollector) Collect(ch chan<- prometheus.Metric) {
	select {
	case metrics := <-c.metricsC:
		for _, m := range metrics {
			ch <- m
		}
	case <-time.After(100 * time.Millisecond):
		c.log.Warn().Msg("could not get metrics fast enough. This is harmless on startup but should not occur shortly after")
	}
}

func (c *SlowCollector) Close() {
	c.stop()
	c.wg.Wait()
}

func (c *SlowCollector) run() {
	defer c.wg.Done()
	defer close(c.metricsC)

	var (
		buffer = make(chan []prometheus.Metric)

		lastDuration     time.Duration
		lastCollect      time.Time
		lastCollectMutex sync.Mutex

		mut sync.Mutex

		wg sync.WaitGroup
	)

	defer wg.Wait()

	defer close(buffer)

	wg.Add(1)
	go func() {
		defer wg.Done()
		var cont = true
		metrics := <-buffer

		for cont {
			select {
			case c.metricsC <- metrics:
				mut.Lock()
				lastCollect = time.Now()
				mut.Unlock()
			case metrics, cont = <-buffer:
			}
		}
	}()

	doCollect := func() []prometheus.Metric {
		var (
			metrics []prometheus.Metric
			ch      = make(chan prometheus.Metric)
		)
		go func() {
			start := time.Now()
			c.Collector.Collect(ch)
			lastDuration = time.Since(start)
			close(ch)
		}()

		for m := range ch {
			metrics = append(metrics, m)
		}
		return metrics
	}

	buffer <- doCollect()

	for {
		var nextCollectDelay time.Duration
		lastCollectMutex.Lock()
		if lastCollect.IsZero() {
			nextCollectDelay = c.delayHint*4 - (lastDuration * 11 / 10)
		} else {
			nextCollectDelay = c.delayHint - (lastDuration * 11 / 10) - time.Since(lastCollect)
		}
		lastCollectMutex.Unlock()

		if nextCollectDelay <= 0 {
			nextCollectDelay = time.Millisecond
		}

		select {
		case <-time.After(nextCollectDelay):
			buffer <- doCollect()

		case <-c.ctx.Done():
			return
		}
	}
}
